#include <stdio.h>
#include "PDF.h"
using namespace std;

char mytolower(char in)
{
  if(in >= 'A' && in <= 'Z')
  {
    int tmp = 'A'-'a';
    in-=tmp;
  }
  return in;
}


void suffixes(char *x, int m, int *suff) {
  int f=0;
  int g=0;
  int i=0;
 
   suff[m - 1] = m;
   g = m - 1;
   for (i = m - 2; i >= 0; --i) {
      if (i > g && suff[i + m - 1 - f] < i - g)
         suff[i] = suff[i + m - 1 - f];
      else {
         if (i < g)
            g = i;
         f = i;
         while (g >= 0 && x[g] == x[g + m - 1 - f])
            --g;
         suff[i] = f - g;
      }
   }
}
 
void preBmGs(char *x, int m, int bmGs[]) {
   int i, j, suff[m];
 
   suffixes(x, m, suff);
 
   for (i = 0; i < m; ++i)
      bmGs[i] = m;
   j = 0;
   for (i = m - 1; i >= 0; --i)
      if (suff[i] == i + 1)
         for (; j < m - 1 - i; ++j)
            if (bmGs[j] == m)
               bmGs[j] = m - 1 - i;
   for (i = 0; i <= m - 2; ++i)
      bmGs[m - 1 - suff[i]] = m - 1 - i;
}


PDF::PDF(char *filename):
  mInitError(false),
  mSrc(NULL),
  mSrcPages(NULL),
  mCurrentPage(NULL),
  mOutline(NULL)
{
    fz_error *error;
 

    error = pdf_newxref(&mSrc);
    if (error)
        die(error);

    error = pdf_loadxref(mSrc, filename);
    if (error)
    {
      fz_warn((char *)"trying to repair");
        error = pdf_repairxref(mSrc, filename);
        if (error)
            die(error);
    }

    error = pdf_decryptxref(mSrc);
    if (error)
        die(error);
}

bool PDF::IsPasswordProtected()
{
    fz_error *error;
     if(mSrc->crypt)
    {
      error = pdf_setpassword(mSrc->crypt, (char *)"");
         if(error)
            return true;
    }
    return false;
}

int PDF::SetPassword(char *password)
{
    fz_error *error;

    if (mSrc->crypt)
    {
        error = pdf_setpassword(mSrc->crypt, password);
        if (error)
            return -1;
    }
    return 0;
}

int PDF::Initialize()
{
    fz_error *error;
   fz_obj *obj;
    error = pdf_loadpagetree(&mSrcPages, mSrc);
    if (error)
        return -1;        

    
    obj = fz_dictgets(mSrc->trailer, (char *)"Root");
    if (!obj)
      {
	die((char *)"Trailer non existant?");
	return -1;
      }
    
    error = pdf_loadindirect(&mSrc->root, mSrc, obj);
    if (error)
        return -1;

    obj = fz_dictgets(mSrc->trailer, (char *)"Info");
    if (obj)
    {
        error = pdf_loadindirect(&mSrc->info, mSrc, obj);
        if (error)
	  return -1;
    }

    error = pdf_loadnametrees(mSrc);
    if (error)
      return -1;

    mOutline = new PDFOutline(this);
    if(mOutline->loadfailure())
    {
        delete mOutline;
        mOutline = NULL;
    }
    return 0;
}

void PDF::die(fz_error *error)
{
    //what do we do???
  printf("We encountered an error\n");
  mInitError = true;
}

void PDF::die(char * error)
{
    //what do we do???
  printf("We encountered an error: %s\n ", error);
  mInitError = true;
}

PDF::~PDF()
{
    if(mSrcPages)
        pdf_droppagetree(mSrcPages);
    mSrcPages = NULL;
    
    if (mSrc->store)
    {
        pdf_dropstore(mSrc->store);
        mSrc->store = (pdf_store_s *)nil;
    }
    pdf_closexref(mSrc);

    if(mOutline)
    {
        delete mOutline;
    }
    mOutline = NULL;
}

int PDF::GetPageCount()
{
    return pdf_getpagecount(mSrcPages);
}

int PDF::NextPage()
{
    if(mCurrentPage == NULL)
    {
        return -1;
    }
    int nextPage = mCurrentPage->GetPageNumber()+1;
    if(nextPage > GetPageCount())
    {
        return -2;
    }
    return JumpToPage(nextPage);
}

int PDF::PreviousPage()
{
    if(mCurrentPage == NULL)
    {
        return -1;
    }
    int nextPage = mCurrentPage->GetPageNumber()-1;
    if(nextPage < 1)
    {
        return -2;
    }
    return JumpToPage(nextPage);
}

int PDF::JumpToPage(int pageNumber)
{
    FreeCurrentPage();
    return SetCurrentPage(LoadPage(pageNumber));
}

PDFPage *PDF::LoadPage(int pageNumber)
{
  if(pageNumber>0 && pageNumber< GetPageCount())
    return new PDFPage(this, pageNumber);
  return NULL;
}

void PDF::DumpAllPagesToText()
{
  for(int i =0; i< GetPageCount() ; i++)
    {
      printf("Page: %d\n",  i+1);
      PDFPage tmp(this, i+1);
      tmp.DebugPageText();
    }
}

void PDF::CaseInsensitiveTextSearch(char *inneedle, int needlelength, int lowpage, int highpage,
				    void (*RecordResult)(int page, int x0, int y0, int x1, int y1, int contextlen, char *context, void *), void *userdata)
{
  int badchar[129];
  int goodsuff[needlelength];
  char needle[needlelength];
  int i;
  //make needle all lowercase
  for(i=0;i<needlelength;i++)
  {
    needle[i] = mytolower(inneedle[i]);
  }
  //initialize the badchar
  for(i=0;i<=128;i++)
    badchar[i] = needlelength;
  for(i=0;i<needlelength-1;i++)
  {
    badchar[(int)needle[i]]=needlelength-i-1;
  }

  //Create the good suffix table
  preBmGs(needle, needlelength, goodsuff);

  for(i=lowpage; i<=highpage;i++)
  {
    PDFPage tmp(this, i);
    tmp.CaseInsensitiveSearchPageText(needle, needlelength, badchar, goodsuff,RecordResult, userdata);
  }  
}

PDFPage *PDF::GetCurrentPage()
{
    return mCurrentPage;
}

void PDF::FreeCurrentPage()
{
    if(mCurrentPage != NULL)
    {
        delete mCurrentPage;
    }
    mCurrentPage = NULL;
}

int PDF::SetCurrentPage(PDFPage *inPage)
{
    if(mCurrentPage != NULL)
    {
        //Are you trying to create a memory leak???
        return -1;
    }
    mCurrentPage = inPage;
    return -1*((int)(inPage == NULL));
}

int PDF::GetLinkPageNumber(pdf_link *link)
{
    int oid = fz_tonum(link->dest);
    int i;
                
    for (i = 0; i < GetPageCount(); i++)
    {
        if (fz_tonum(mSrcPages->pref[i]) == oid)
        {
            return i+1;
        }
    }
    return -2;
}

int PDF::HasLinkAtLocation(int x, int y, char *buf, bool getPageNumber)
{
    pdf_link *link = NULL;
    int counter=0;
    if(mCurrentPage->Loaded())
    {
        x+=(int)mCurrentPage->mPage->mediabox.x0;
        y+=(int)mCurrentPage->mPage->mediabox.y0;
    

        for (link = mCurrentPage->mPage->links; link; link = link->next)
        {
            counter++;
            if (x >= link->rect.x0 && x <= link->rect.x1)
                if (y >= link->rect.y0 && y <= link->rect.y1)
                {
                    break;
                }
        }
    }

    if (link)
    {
        if(getPageNumber)
        {
            if (fz_isstring(link->dest))
            {                
                memcpy(buf, fz_tostrbuf(link->dest), fz_tostrlen(link->dest));
                buf[fz_tostrlen(link->dest)] = 0;
                return 0;
            }
            if (fz_isindirect(link->dest))
            {
                return GetLinkPageNumber(link);
            }
            printf("We didn't find the page for this link!\n");
        }
        else
        {
            return 0;
        }
    }
    else
    {
        return -1;
    }
    return -2;
}

PDFOutline *PDF::GetOutline()
{
    return mOutline;
}
